/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     | Website:  https://openfoam.org
    \\  /    A nd           | Copyright (C) 2011-2020 OpenFOAM Foundation
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "volFieldValue.H"
#include "volFields.H"

// * * * * * * * * * * * * Protected Member Functions  * * * * * * * * * * * //

template<class Type>
bool Foam::functionObjects::fieldValues::volFieldValue::validField
(
    const word& fieldName
) const
{
    typedef GeometricField<Type, fvPatchField, volMesh> vf;

    if (obr_.foundObject<vf>(fieldName))
    {
        return true;
    }

    return false;
}


template<class Type>
Foam::tmp<Foam::Field<Type>>
Foam::functionObjects::fieldValues::volFieldValue::setFieldValues
(
    const word& fieldName,
    const bool mustGet
) const
{
    typedef GeometricField<Type, fvPatchField, volMesh> vf;

    if (obr_.foundObject<vf>(fieldName))
    {
        return filterField(obr_.lookupObject<vf>(fieldName));
    }

    if (mustGet)
    {
        FatalErrorInFunction
            << "Field " << fieldName << " not found in database"
            << abort(FatalError);
    }

    return tmp<Field<Type>>(new Field<Type>(0.0));
}


template<class Type>
Type Foam::functionObjects::fieldValues::volFieldValue::processValues
(
    const Field<Type>& values,
    const scalarField& V,
    const scalarField& weightField
) const
{
    Type result = Zero;
    switch (operation_)
    {
        case operationType::sum:
        {
            result = gSum(values);
            break;
        }
        case operationType::weightedSum:
        {
            result = gSum(weightField*values);
            break;
        }
        case operationType::sumMag:
        {
            result = gSum(cmptMag(values));
            break;
        }
        case operationType::average:
        {
            result = gSum(values)/nCells();
            break;
        }
        case operationType::weightedAverage:
        {
            result = gSum(weightField*values)/max(gSum(weightField), vSmall);
            break;
        }
        case operationType::volAverage:
        {
            result = gSum(V*values)/this->V();
            break;
        }
        case operationType::weightedVolAverage:
        {
            result =
                gSum(weightField*V*values)/max(gSum(weightField*V), vSmall);
            break;
        }
        case operationType::volIntegrate:
        {
            result = gSum(V*values);
            break;
        }
        case operationType::weightedVolIntegrate:
        {
            result = gSum(weightField*V*values);
            break;
        }
        case operationType::min:
        {
            result = gMin(values);
            break;
        }
        case operationType::max:
        {
            result = gMax(values);
            break;
        }
        case operationType::CoV:
        {
            Type meanValue = gSum(values*V)/this->V();

            const label nComp = pTraits<Type>::nComponents;

            for (direction d=0; d<nComp; ++d)
            {
                scalarField vals(values.component(d));
                scalar mean = component(meanValue, d);
                scalar& res = setComponent(result, d);

                res = sqrt(gSum(V*sqr(vals - mean))/this->V())/mean;
            }

            break;
        }
        case operationType::none:
        {}
    }

    return result;
}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

template<class Type>
bool Foam::functionObjects::fieldValues::volFieldValue::writeValues
(
    const word& fieldName
)
{
    const bool ok = validField<Type>(fieldName);

    if (ok)
    {
        Field<Type> values(setFieldValues<Type>(fieldName));
        scalarField V(filterField(fieldValue::mesh_.V()));
        scalarField weightField(values.size(), 1.0);

        forAll(weightFieldNames_, i)
        {
            weightField *= setFieldValues<scalar>(weightFieldNames_[i], true);
        }

        Type result = processValues(values, V, weightField);

        if (Pstream::master())
        {
            // Add to result dictionary, over-writing any previous entry
            resultDict_.add(fieldName, result, true);

            file()<< tab << result;

            Log << "    " << operationTypeNames_[operation_]
                << "(" << volRegion::regionName_ << ") of " << fieldName
                <<  " = " << result << endl;
        }

        if (writeFields_)
        {
            IOField<Type>
            (
                IOobject
                (
                    fieldName + '_' + regionTypeNames_[regionType_]
                  + '-' + volRegion::regionName_,
                    obr_.time().timeName(),
                    obr_,
                    IOobject::NO_READ,
                    IOobject::NO_WRITE
                ),
                (weightField*values).ref()
            ).write();
        }
    }

    return ok;
}


template<class Type>
Foam::tmp<Foam::Field<Type>>
Foam::functionObjects::fieldValues::volFieldValue::filterField
(
    const Field<Type>& field
) const
{
    if (isNull(cellIDs()))
    {
        return field;
    }
    else
    {
        return tmp<Field<Type>>(new Field<Type>(field, cellIDs()));
    }
}


// ************************************************************************* //
